import logging

import click

from .http_service.app import app
from .scenarios import HttpScenario, TelnetScenario
from .service_mocks.http_service_mock import WerkzeugBasedHttpService
from .telnet_service.protocol import TelnetServerClientProtocol
from .telnet_service.telnet_server import TelnetServer

logger = logging.getLogger(__name__)


@click.group()
def cli():
    pass


@cli.command("http")
@click.option("--port",
              default=8080,
              show_default=True,
              help="HTTP server port.")
@click.option("--scenario",
              required=True,
              type=click.Choice(HttpScenario.names()),
              help="HTTP server behaviour.")
def run_http_server(scenario, port):
    logger.debug("Starting `http` server...")
    app.config.update(
        SCENARIO=HttpScenario[scenario],
    )
    WerkzeugBasedHttpService(
        app=app,
        host=f"127.0.0.1",
        port=port,
    ).start()
    logger.debug(f"`http` server has been started on port {port}.")


@cli.command("https")
@click.option("--port",
              default=8443,
              show_default=True,
              help="HTTPS server port.")
@click.option("--scenario",
              required=True,
              type=click.Choice(HttpScenario.names()),
              help="HTTP server behaviour.")
def run_https_server(scenario, port):
    logger.debug("Starting `https` server...")
    app.config.update(
        SCENARIO=HttpScenario[scenario],
    )
    WerkzeugBasedHttpService(
        app=app,
        host=f"127.0.0.1",
        port=port,
        ssl=True,
    ).start()
    logger.debug(f"`https` server has been started on port {port}.")


@cli.command("telnet")
@click.option("--port",
              default=8023,
              show_default=True,
              help="Telnet server port.")
@click.option("--scenario",
              required=True,
              type=click.Choice(TelnetScenario.names()),
              help="Telnet server behaviour.")
def run_telnet_server(scenario, port):
    logger.debug("Starting `telnet` server...")
    TelnetServer(
        host="127.0.0.1",
        port=port,
        protocol=lambda: TelnetServerClientProtocol(TelnetScenario[scenario]),
    ).run()
    logger.debug(f"`telnet` server has been started on port {port}.")



import logging

from ..http_service.app import app
from ..scenarios import HttpScenario
from .base_http_service import WerkzeugBasedHttpService

logger = logging.getLogger(__name__)


class HttpScenarioService(WerkzeugBasedHttpService):
    def __init__(self, host: str, port: int, scenario: HttpScenario):
        app.config.update(SCENARIO=scenario)
        super().__init__(host, port, app)



import asyncio
import logging

logger = logging.getLogger(__name__)


class TelnetServer:
    def __init__(self, host: str, port: int, protocol):
        self.loop = asyncio.get_event_loop()

        coro = self.loop.create_server(protocol, host, port)
        self.server = self.loop.run_until_complete(coro)

    def run(self):
        logger.debug(f"Serving on {self.server.sockets[0].getsockname()}")
        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            self.server.close()
            self.loop.run_until_complete(self.server.wait_closed())



import logging
import socket
import time

logger = logging.getLogger(__name__)


class BaseService:

    socket_type = socket.SOCK_STREAM

    def __init__(self, host: str, port: int):
        self.host = host
        self.port, self.dibbed_port_socket = self.dib_port(port)

    def _wait_for_service(self):
        elapsed_time = 0
        start_time = time.time()
        while elapsed_time < 5:
            s = socket.socket(type=self.socket_type)
            s.settimeout(1)
            try:
                s.connect((self.host, self.port))
            except (ConnectionRefusedError, ConnectionAbortedError,
                    socket.timeout):
                elapsed_time = time.time() - start_time
                s.close()
            else:
                s.close()
                break
        else:
            raise TimeoutError(f"{self.__class__.__name__} "
                               f"couldn't be set up before test.")

    def start(self):
        raise NotImplementedError()

    def teardown(self):
        raise NotImplementedError()

    def __enter__(self):
        logger.debug(f"Starting {self}...")
        self.dibbed_port_socket.close()
        self.start()
        self._wait_for_service()
        logger.debug(f"{self} has been started.")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.debug(f"Terminating {self}...")
        self.teardown()
        logger.debug(f"{self} has been terminated.")

    @staticmethod
    def dib_port(port: int = 0) -> (int, socket.socket):
        socket_ = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        socket_.bind(("", port))
        return int(socket_.getsockname()[1]), socket_

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(host='{self.host}', port={self.port})"
        )



from logging import getLogger
import socketserver
from typing import Pattern, Union
from unittest import mock

logger = getLogger(__name__)


class UDPServer(socketserver.ThreadingUDPServer):
    allow_reuse_address = True

    def __init__(
            self,
            server_address,
            request_handler_class,
            bind_and_activate=True,
    ):
        super().__init__(
            server_address, request_handler_class, bind_and_activate,
        )
        self.handlers = {}

    def get_handler(
            self,
            command: Union[bytes, Pattern[bytes]],
    ) -> mock.MagicMock:
        handler = self.handlers.get(command)
        if handler:
            return handler

        for pattern_key in self.handlers:
            if isinstance(pattern_key, Pattern):
                if pattern_key.match(command):
                    return self.handlers[pattern_key]

        handler = mock.MagicMock(name="default_handler")
        handler.return_value = b""

        return handler

    def get_command_mock(
            self,
            command: Union[bytes, Pattern[bytes]],
    ) -> mock.MagicMock:
        mocked_handler = mock.MagicMock(name=command)
        self.handlers[command] = mocked_handler
        return mocked_handler


class UDPHandler(socketserver.DatagramRequestHandler):
    def handle(self):
        data = self.rfile.read()
        handler = self.server.get_handler(data)
        self.wfile.write(handler())



import logging
from unittest import mock

from flask import Flask

from .base_http_service import WerkzeugBasedHttpService

logger = logging.getLogger(__name__)


class HttpServiceMock(WerkzeugBasedHttpService):
    def __init__(self, host: str, port: int, ssl=False):
        super().__init__(host, port, Flask("target"), ssl)

    def get_route_mock(self, rule, **options):
        mocked_view = mock.AsyncMock(name=rule, spec=lambda: None)
        self.app.add_url_rule(rule,
                              endpoint=rule,
                              view_func=mocked_view,
                              **options)
        logger.debug(f"{self} mock for '{rule}' has been added.")
        return mocked_view



from pathlib import Path
import tempfile

from OpenSSL import crypto
from gunicorn.app.base import BaseApplication
from werkzeug.serving import generate_adhoc_ssl_pair


class GunicornServer(BaseApplication):
    def __init__(self, app, **kwargs):
        self.options = kwargs
        self.application = app
        super().__init__()

    def load_config(self):
        if self.options.get("ssl"):
            cert_path, pkey_path = self.generate_devel_ssl_pair()
            self.options["certfile"] = str(cert_path)
            self.options["keyfile"] = str(pkey_path)

        config = {
            key: value for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application

    @staticmethod
    def generate_devel_ssl_pair() -> (Path, Path):
        cert_path = Path(tempfile.gettempdir()) / "threat9-test-bed.crt"
        pkey_path = Path(tempfile.gettempdir()) / "threat9-test-bed.key"

        if not cert_path.exists() or not pkey_path.exists():
            cert, pkey = generate_adhoc_ssl_pair()
            with open(cert_path, "wb") as f:
                f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
            with open(pkey_path, "wb") as f:
                f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))

        return cert_path, pkey_path



from enum import Enum, unique


@unique
class HttpScenario(Enum):
    EMPTY_RESPONSE = "empty_response"
    TRASH = "trash"
    NOT_FOUND = "not_found"
    FOUND = "found"
    REDIRECT = "redirect"
    TIMEOUT = "timeout"
    ERROR = "error"

    @staticmethod
    def names():
        return [element.name for element in HttpScenario]


@unique
class TelnetScenario(Enum):
    GENERIC = "generic"
    AUTHORIZED = "authorized"
    NOT_AUTHORIZED = "not_authorized"
    TIMEOUT = "timeout"

    @staticmethod
    def names():
        return [element.name for element in TelnetScenario]



from .http_scenario_service import HttpScenarioService  # noqa: F401
from .http_service_mock import HttpServiceMock  # noqa: F401
from .tcp_service_mock import TCPServiceMock  # noqa: F401
from .telnet_service_mock import TelnetServiceMock  # noqa: F401
from .udp_service_mock import UDPServiceMock  # noqa: F401



import logging
import time

from faker import Faker
from flask import Flask, abort, g, redirect

from ..scenarios import HttpScenario

logger = logging.getLogger(__name__)


def get_faker():
    faker = g.get("faker", None)
    if faker is None:
        faker = Faker()
        g.user = faker
    return faker


app = Flask(__name__)

ALLOWED_METHODS = [
    "GET",
    "POST",
    "PATCH",
    "PUT",
    "OPTIONS",
    "DELETE",
]


@app.route("/", defaults={"path": ""}, methods=ALLOWED_METHODS)
@app.route("/<path:path>", methods=ALLOWED_METHODS)
def catch_all(path):
    scenario_handler = SCENARIO_TO_HANDLER_MAP.get(
        app.config["SCENARIO"],
        error,
    )
    logger.debug(
        f"Executing '{scenario_handler.__name__}' scenario handler..."
    )
    return scenario_handler()


def empty_response():
    return "", 200


def trash():
    return get_faker().paragraph(variable_nb_sentences=True), 200


def not_found():
    abort(404)


def found():
    return "OK", 200


def redirect_():
    return redirect("https://www.youtube.com/watch?v=dQw4w9WgXcQ")


def timeout():
    time.sleep(60 * 60)
    return found()


def error():
    abort(500)


SCENARIO_TO_HANDLER_MAP = {
    HttpScenario.EMPTY_RESPONSE: empty_response,
    HttpScenario.TRASH: trash,
    HttpScenario.NOT_FOUND: not_found,
    HttpScenario.FOUND: found,
    HttpScenario.REDIRECT: redirect_,
    HttpScenario.TIMEOUT: timeout,
    HttpScenario.ERROR: error,
}



from logging import getLogger
import threading
from typing import Pattern, Union
from unittest import mock

from ..tcp_service.tcp_server import TCPHandler, TCPServer
from .base_service import BaseService

logger = getLogger(__name__)


class TCPServiceMock(BaseService):

    def __init__(self, host: str, port: int):
        super().__init__(host, port)
        self.server = TCPServer((self.host, self.port), TCPHandler, False)
        self.server_thread = threading.Thread(target=self.server.serve_forever)

    def start(self):
        self.server.server_bind()
        self.server.server_activate()
        self.server_thread.start()

    def teardown(self):
        self.server.shutdown()
        self.server_thread.join()
        self.server.server_close()

    def get_command_mock(
            self,
            command: Union[bytes, Pattern[bytes]],
    ) -> mock.MagicMock:
        logger.debug(f"{self} mock for '{command}' has been added.")
        return self.server.get_command_mock(command)



import logging
import multiprocessing
import threading
from wsgiref.simple_server import make_server as wsgiref_make_server

from flask import Flask
from werkzeug.serving import make_server as werkzeug_make_server

from ..http_service.gunicorn_server import GunicornServer
from .base_service import BaseService

logger = logging.getLogger(__name__)


class GunicornBasedHttpService(BaseService):
    """ `gunicorn` based HTTP service

    `Flask` application served using `gunicorn` in separate process using
    async workers (threads in this case).

    Application served by this base class suppose to handle unbuffered
    requests, `nginx` in this case is no option hence async workers.
    """
    def __init__(self, host: str, port: int, app: Flask, ssl=False):
        super().__init__(host, port)
        self.app = app
        self.server = GunicornServer(
            app=self.app,
            bind=f"{self.host}:{self.port}",
            worker_class="gthread",
            threads=8,
            ssl=ssl,
            accesslog="-",
        )
        self.server_process = multiprocessing.Process(target=self.server.run)

    def start(self):
        self.server_process.start()

    def teardown(self):
        self.server_process.terminate()
        self.server_process.join()


class WSGIRefBasedHttpService(BaseService):
    """ `wsgiref` based HTTP service

    `Flask` application served using `wsgiref` in separate thread.

    We can leverage shared state between main thread and thread handling
    `wsgiref` server and dynamically attach `Mock` object as view functions.
    """
    def __init__(self, host: str, port: int, app: Flask):
        super().__init__(host, port)
        self.app = app
        self.server = wsgiref_make_server(self.host, self.port, self.app)
        self.server_thread = threading.Thread(target=self.server.serve_forever)

    def start(self):
        self.server_thread.start()

    def teardown(self):
        self.server.shutdown()
        self.server_thread.join()
        self.server.server_close()


class WerkzeugBasedHttpService(BaseService):
    """ `werkzeug` based HTTP service

    `Flask` application served using `werkzeug` in separate thread.

    We can leverage shared state between main thread and thread handling
    `wsgiref` server and dynamically attach `Mock` object as view functions.
    """
    def __init__(self, host: str, port: int, app: Flask, ssl=False):
        super().__init__(host, port)
        self.app = app
        self.dibbed_port_socket.close()  # werkzeug binds port on server init
        self.server = werkzeug_make_server(
            self.host, self.port, self.app,
            threaded=True,
            ssl_context="adhoc" if ssl else None
        )
        self.server_thread = threading.Thread(target=self.server.serve_forever)

    def start(self):
        self.server_thread.start()

    def teardown(self):
        self.server.shutdown()
        self.server_thread.join()
        self.server.server_close()



import asyncio
import logging
import typing

from faker import Faker

from threat9_test_bed.scenarios import TelnetScenario

logger = logging.getLogger(__name__)
faker = Faker()


def authorized(func):
    def _wrapper(self, data):
        message = data.decode().strip()
        if not self.authorized:
            if not self.login:
                self.login = message
                self.transport.write(b"Password: ")
                return
            else:
                self.password = message

            if (self.login, self.password) in self.creds:
                self.authorized = True
                self.transport.write(self.prompt.encode())
            else:
                self.transport.write(b"\r\nLogin incorrect\r\ntarget login: ")
                self.login = None
                self.password = None
            return
        else:
            func(self, data)

    return _wrapper


class GreedyList(list):
    def __contains__(self, item):
        return True


class TelnetServerClientProtocol(asyncio.Protocol):
    def __init__(self, scenario: TelnetScenario):
        self.transport = None
        self.remote_address = None
        self.scenario = scenario

        self.login = None
        self.password = None
        self.authorized = False

        self.banner = b""
        self._command_mocks = {}
        self._creds = [
            ("admin", "admin"),
            ("kocia", "dupa"),
        ]

    @property
    def creds(self):
        if self.scenario is TelnetScenario.NOT_AUTHORIZED:
            return []
        elif self.scenario is TelnetScenario.AUTHORIZED:
            return GreedyList()
        elif self.scenario is TelnetScenario.GENERIC:
            return self._creds

    @property
    def prompt(self):
        return f"{self.login}@target:~$ "

    def connection_made(self, transport: asyncio.Transport):
        self.remote_address = transport.get_extra_info("peername")
        logger.debug(f"Connection from {self.remote_address}")
        self.transport = transport
        if self.banner:
            self.transport.write(self.banner + b"\r\n")
        self.transport.write(b"Login: ")

    def _get_handler(
            self,
            command: typing.Union[str, typing.Pattern[str]],
    ) -> typing.Callable:

        handler = self._command_mocks.get(command)
        if handler:
            return handler

        for pattern_key in self._command_mocks:
            if isinstance(pattern_key, typing.Pattern):
                if pattern_key.match(command):
                    return self._command_mocks[pattern_key]

        return faker.paragraph

    @authorized
    def data_received(self, data: bytes):
        logger.debug(f"{self.remote_address} send: {data}")
        command = data.decode().strip()
        handler = self._get_handler(command)
        self.transport.write(
            f"{handler()}\r\n"f"{self.prompt}".encode()
        )

    def add_command_handler(self, command: str, handler: typing.Callable):
        self._command_mocks[command] = handler

    def add_credentials(self, login: str, password: str):
        self._creds.append((login, password))

    def add_banner(self, banner: bytes):
        self.banner = banner



import logging
import threading
import typing
from unittest import mock

from ..scenarios import TelnetScenario
from ..telnet_service.protocol import TelnetServerClientProtocol
from ..telnet_service.telnet_server import TelnetServer
from .base_service import BaseService

logger = logging.getLogger(__name__)


class TelnetServiceMock(BaseService):
    def __init__(self,
                 host: str, port: int,
                 scenario: TelnetScenario=TelnetScenario.AUTHORIZED):
        super().__init__(host, port)
        self.protocol = TelnetServerClientProtocol(scenario)
        self.server = TelnetServer(
            self.host,
            self.port,
            lambda: self.protocol
        )
        self.server_thread = threading.Thread(target=self.server.run)

    def start(self):
        self.server_thread.start()

    def teardown(self):
        self.server.loop.call_soon_threadsafe(self.server.loop.stop)
        self.server_thread.join()

    def get_command_mock(
            self,
            command: typing.Union[str, typing.Pattern[str]],
    ):
        command_mock = mock.MagicMock(name=command)
        self.protocol.add_command_handler(command, command_mock)
        return command_mock

    def add_credentials(self, login: str, password: str):
        """ Add custom credentials pair. """
        self.protocol.add_credentials(login, password)

    def add_banner(self, banner: bytes):
        """ Add welcoming banner after connection. """
        self.protocol.add_banner(banner)
